/*
 *  pong.cpp
 *  This is a simple double buffered program.
 *  Use the left and right arrow keys to move the paddle.
 *  Use the home key to re-center the paddle and return the ball "home."
 *  The beginnings of the classic video game.  I say "The beginnings"
 *  because this program doesn't perform any collision detection or response.
 */


#include <unistd.h>
#include "Angel.h"


const int SLICES = 72;  // For generating the circle's vertices.


/* When drawing, we need to switch between the VAO for the paddle and the
 * ball, hence we need global variables for holding the VAO handles.
 */

vec3 paddlePos(0.0, -47.5, 0.0);      // Initial position.
vec3 paddleVel(5.0, 0.0, 0.0);        /* Magnitude of velocity when moving
                                       * the paddle left or right.
                                       */
vec4 paddleCol(1.0, 1.0, 1.0, 1.0);   // Color.
GLuint paddleVAO;


const vec3 START_POS(0.0, 0.0, 0.0);
const vec3 START_VEL(0.1, 0.5, 0.0);
vec3 ballPos = START_POS;
vec3 ballVel = START_VEL;
vec4 ballCol(1.0, 1.0, 1.0, 1.0);
GLuint ballVAO;


/* Handles for uniform variables in the vertex shader. */

GLuint projection;
GLuint model_view;
GLuint vColor;


/* Creates a VAO and buffers vertices for a circle centered at the origin
 * in the z = 0.0 plane.  loc is the handle for vPosition in the vertex
 * shader.  Returns the VAO handle.
 */

GLuint createCircle(GLfloat radius, GLuint loc)
{
   GLuint vao, buffer;

   vec3 points[SLICES];
   GLfloat angle = 0.0;
   GLfloat sliceAngle = 2.0 * M_PI / (GLfloat) SLICES;

   for (int i = 0; i < SLICES; i++)
   {
      points[i] = radius * vec3(cos(angle), sin(angle), 0.0);
      angle += sliceAngle;
   }

   glGenVertexArrays(1, &vao);
   glBindVertexArray(vao);

   glGenBuffers(1, &buffer);
   glBindBuffer(GL_ARRAY_BUFFER, buffer);

   glBufferData(GL_ARRAY_BUFFER, sizeof(points),
                points, GL_STATIC_DRAW);

   glEnableVertexAttribArray(loc);
   glVertexAttribPointer(loc, 3, GL_FLOAT, GL_FALSE, 0,
                         BUFFER_OFFSET(0));

   return vao;
}


void display(void)
{
   mat4 mv;   /* Model view matrix */

   glClear(GL_COLOR_BUFFER_BIT);

   // Draw the paddle.

   glBindVertexArray(paddleVAO);
   glUniform4fv(vColor, 1, paddleCol);
   mv = Translate(paddlePos);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, mv);
   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

   // Draw the ball.

   glBindVertexArray(ballVAO);
   glUniform4fv(vColor, 1, ballCol);
   mv = Translate(ballPos);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, mv);
   glDrawArrays(GL_TRIANGLE_FAN, 0, SLICES);
   
   glutSwapBuffers();
}


void init(void) 
{
   /* Vertices for the paddle, centered at the origin.  Ultimately,
    * these will be buffered, hence we use a local array.
    */

   vec3 points[4] = { vec3(-10.0, -2.5, 0.0), vec3(10.0, -2.5, 0.0), 
                      vec3(10.0, 2.5, 0.0), vec3(-10.0, 2.5, 0.0)
                    };

   glClearColor (0.0, 0.0, 0.0, 1.0);
   glShadeModel (GL_FLAT);

   /* Compile, link, and bind the shader programs. */

   GLuint program = InitShader("vshader.glsl", "fshader.glsl");
   glUseProgram(program);

   /* Get the locations of the uniform variables in the vertex shader.
    * Because we'll be updating these values throughout the execution of
    * the program, these variables are all global.  Because we're not
    * loading differing projection matrices into the vertex shader,
    * we really don't need a global for the projection variable, but
    * setting it up this way would allow us to adjust to aspect ratio
    * changes at a later time.
    */

   vColor = glGetUniformLocation(program, "vColor");
   model_view = glGetUniformLocation(program, "model_view");
   projection = glGetUniformLocation(program, "projection");

   // Get the location of the vertex shader's vPosition input.

   GLuint vPosition = glGetAttribLocation(program, "vPosition");

   // Create a VAO for the paddle and buffer its vertices.

   glGenVertexArrays(1, &paddleVAO);
   glBindVertexArray(paddleVAO);

   GLuint buffer;
   glGenBuffers(1, &buffer);
   glBindBuffer(GL_ARRAY_BUFFER, buffer);
   glBufferData(GL_ARRAY_BUFFER, sizeof(points), points, GL_STATIC_DRAW);

   glEnableVertexAttribArray(vPosition);
   glVertexAttribPointer(vPosition, 3, GL_FLOAT, GL_FALSE, 0,
                         BUFFER_OFFSET(0));

   /* Create a VAO for the ball, buffering its vertices. */

   ballVAO = createCircle(5.0, vPosition);
}


void reshape(int w, int h)
{
   mat4 p;    /* Projection matrix */

   glViewport (0, 0, (GLsizei) w, (GLsizei) h);

   p = Ortho(-50.0, 50.0, -50.0, 50.0, -1.0, 1.0);
   glUniformMatrix4fv(projection, 1, GL_TRUE, p);
}


void keyboard(int key, int x, int y)
{
   // Limited here by keyboard repeat rate.

	switch (key)
	{
	case GLUT_KEY_LEFT:
		paddlePos -= paddleVel;
		break;

	case GLUT_KEY_RIGHT:
		paddlePos += paddleVel;
		break;

	case GLUT_KEY_HOME:
		paddlePos = vec3(0.0, -47.5, 0.0);
		ballPos = START_POS;
      ballVel = START_VEL;
		break;

   default:
      break;
	}

   /* Presumably, we've moved the paddle, so redraw the scene. */

	glutPostRedisplay();
}


/* Check for collisions, and respond, and update the ball's position.
 * Collision response is naive --- simply negate the affected velocity
 * component.  When the paddle is struck, the velocity is increased by
 * 15%, and then the velocity's x component is increased by another 10%.
 */

void idle(void)
{
   /* Handle paddle collisions. */
   if (ballPos.y - 5.0 <= -45.0
       && ballPos.x >= paddlePos.x - 10.0
       && ballPos.x <= paddlePos.x + 10.0)
   {
      ballVel.y = -ballVel.y;
      ballVel *= 1.15;
      ballVel.x *= 1.1;
   }

   /* Handle collisions with side walls. */

   if (ballPos.x + 5.0 >= 50.0 || ballPos.x - 5.0 <= -50.0)
      ballVel.x = -ballVel.x;

   /* Handle collisions with top wall. */

   if (ballPos.y + 5.0 >= 50.0)
      ballVel.y = -ballVel.y;

   /* Handle the ball exiting the field of play via the screen bottom. */

   if (ballPos.y + 5.0 <= -50.0)
   {
#ifdef WIN32
      Sleep(1000);
#else
      sleep(1);
#endif
      ballPos = START_POS;
      ballVel = START_VEL;
   }

   ballPos += ballVel;
	glutPostRedisplay();
}

   
/* 
 *  Request double buffer display mode.
 *  Register "special" input callback functions for arrow and home keys.
 */

int main(int argc, char** argv)
{
   glutInit(&argc, argv);
   glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB);
   glutInitWindowSize (500, 500); 
   glutInitWindowPosition (100, 100);
   glutInitContextVersion(3, 2);
   glutInitContextProfile(GLUT_CORE_PROFILE);
   glutCreateWindow ("Pong");

   glewExperimental = GL_TRUE;
   glewInit();

   init ();

   glutDisplayFunc(display); 
   glutReshapeFunc(reshape); 
   glutSpecialFunc(keyboard);
   glutIdleFunc(idle);

   glutMainLoop();

   return 0;
}
